unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, Buttons;

const
  formStr = 't%d: %6.2f s: %7.2f  v: %5.2f  a: %5.2f';

type
  tdyn = record
    t,          // Gesamtzeit
    dt,         // Zeitschritt
    s,          // Gesamtstrecke
    v,          // momentane Geschwindigkeit
    a: double;  // vorgegebene Beschleunigung
  end;


  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Image1: TImage;
    Memo1: TMemo;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);

  private
  const
    x_scale = 770 / 450;//Pixel pro Sekunde
    y_scale_s = -150 / 6200;
    y_scale_v = -4;
    y_scale_a = -40;
    y_s_Offset = 150;
    y_v_Offset = 300;
    y_a_Offset = 450;
    col_s = clblue;
    col_v = clgreen;
    col_a = clred;

  var
    fdyn1, fdyn2: tdyn;

    { private declarations }
    procedure Speedtest;
    procedure FdynInit(Sender: TObject);
    procedure AusgabeDyn(Sender: TObject);
    function beschleunigungszeit(v0, Bschl, brems, s_ges: double): double;
  public
    { public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}
function BremsWeg(const dyn: tdyn;Beschl:double):double;
begin
  IF Beschl*dyn.v >= 0 then
    result := 1e308
  else
     with dyn do
       result := ABS(v*v/2*Beschl);
end;

function BremsZeit(const dyn: tdyn;Beschl:double):double;
begin
  IF Beschl*dyn.v >= 0 then
    result := 1e308
  else
    with dyn do
     result := ABS(v/Beschl);
end;

procedure NextTimeStep(const dyn: tdyn; out res: tdyn);
// Einen Zeitschritt mit den vorgegebenen Werten berechnen
var
  ds, tmp: double;

  procedure FahrtRichtungswechsel;
  // Wechsel der Fahrtrichtung bedeutet auch
  // zu Beginn Vorzeichen v <>  Vorzeichen a
  begin
    with res do
    begin
      //t_br  := -v/a;    //Bremszeit
      //t_be:= dt-t_br = dt+v/a; // Verbleibende Beschleunigungszeit
      ds := v/a;        //nur einmal v/a rechnen
      // Verbleibende Beschleunigungszeit tmp = Zeit
      tmp := dt + ds;
      // 0.5*v*v/a+0.5*a*sqr(tmp); //Bremsweg-Beschleunigungsweg
      //Bremsweg in die Richtung von v Beschleunigung in Gegenrichtung
      ds := (v * ds - sqr(tmp) * a) * 0.5;
      // ab hier tmp = v
      tmp := a * tmp;
    end;
  end;

begin
  //Nur Kopieren, wenn Eingangs<> Ausgangsvariable.
  if PtrUInt(@dyn) <> PtrUInt(@res) then
    res := dyn;

  with res do
  begin
    // Keine Zeitaenderung-> keine Aenderung
    if dt <= 0 then
      EXIT;

    t := t + dt;
    // Keine Beschleunigung -> weiter wie bisher v= const a= 0;
    if a = 0 then begin
       s := s+v*dt;
       EXIT;
       end;

    // Beschleunigte Bewegung. Annahme keine Richtungsumkehr
    // Streckenabschnitt fuer diesen Zeitschritt
    ds := (a * dt * 0.5 + v) * dt;
    //neue Geschwindigkeit
    tmp :=  a * dt+ v;
    // Start und Endgeschwindigkeit in selber Richtung geblieben?
    if v * tmp < 0 then
      FahrtRichtungswechsel;
    s := ds + s;
    v := tmp;
  end;
end;

function TForm1.Beschleunigungszeit(v0, Bschl, brems, s_ges: double): double;
{Bestimmt bei vorgegener Strecke die Zeit, die man noch mit Beschl beschleunigen
(um noch passend bremsend mit BremsBeschleunigung zum Stehen zu kommen}
var
  fak2, p2, q: double;
begin
  //Reicht der Weg als Bremsweg
  if ABS(2 * brems * s_ges) < sqr(v0) then
  begin
    Result := -1;
    exit;
  end;
  //0= Beschleunigungsweg + Bremsweg-s_ges
  //Beschleunigungsweg = v0*t+0.5*Bschl*sqr(t)
  //Bremsweg = (v0+Bschl*t)^2/(2*brems)
  //Nach t umstellen und aufloesen
  If brems=Bschl then begin
    fak2 :=  1/Bschl;
    p2 := v0 * fak2 * 2;
    q := (sqr(v0) / (2 * Bschl) - s_ges) * fak2;
    end
  else begin
    fak2 := (sqr(Bschl) / brems + Bschl) * 0.5;
    p2 := ((1 + Bschl / brems) * v0) / (fak2 * 2);
    q := (sqr(v0) / (2 * brems) - s_ges) / fak2;
    end;
  //x_12 := -p/2+-sqrt((-p/2)²-q)
  Result := -p2 + sqrt(sqr(p2) - q);
end;

procedure TForm1.FdynInit(Sender: TObject);
begin
  with image1.Canvas do
  begin
    brush.style := bsSolid;
    brush.color := RGBtoColor(160, 160, 160);
    FillRect(0, 0, Width, Height);
    image1.refresh;

    Pen.Color := clBlack;
    Line(0, y_s_Offset, Width, y_s_Offset);
    Line(0, y_v_Offset, Width, y_v_Offset);
    Line(0, y_a_Offset, Width, y_a_Offset);
  end;
  image1.Update;
  Form1.Refresh;

  with fDyn1 do
  begin
    t := 0;
    dt := 1/x_scale; //horizontal eine Berechnung pro Punkt.
    s := 0;
    v := 0;
    a := 0;
  end;
end;

procedure Tform1.AusgabeDyn(Sender: TObject);
var
  cv: tCanvas;
  x: integer;
begin
  cv := image1.canvas;
  with fDyn1 do
  begin
    x := trunc(x_scale * t + 0.5);
    cv.Pixels[x, round(s * y_scale_s) + y_s_Offset] := col_s;
    cv.Pixels[x, round(v * y_scale_v) + y_v_Offset] := col_v;
    cv.Pixels[x, round(a * y_scale_a) + y_a_Offset] := col_a;
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);

var
  te, s0, s_ges: double;
  idx: integer;

  procedure Berechnung(dt, da: double);
  //Ausgabe der Werte
  var
    deltaT : double;
    x: integer;
  begin
    Inc(idx);
    deltaT := Fdyn1.dt;
    x := trunc(dt/deltaT)+1;
    Fdyn1.dt := dt/x;
    // Beschleunigung ändert sich nicht
    if da = 0.0 then
      repeat
        NextTimeStep(Fdyn1, Fdyn1);
        AusgabeDyn(self);
        dec(x);
      until x<=0
      // Beschleunigung ändert sich
    else begin
      da := da * fdyn1.dt;
      repeat
        NextTimeStep(Fdyn1, Fdyn1);
        fDyn1.a := fDyn1.a + da;
        AusgabeDyn(self);
        dec(x);
      until x<=0
      end;
    Fdyn1.dt := deltaT;
    with fDyn1 do
    begin
      x := trunc(x_scale * t + 0.5);
      memo1.Lines.Add(Format(FormStr, [idx, t, s, v, a]));
    end;
    image1.canvas.Line(x, 0, x, Height);
  end;

begin
  memo1.Clear;
  Button1.Visible := False;
  FdynInit(self);
  idx := 0;
  with fdyn1 do
    memo1.Lines.Add(Format(FormStr, [idx, t, s, v, a]));
  //Beschleunung in 1 Sekunde um 0.1 erhöhen
  te := 1;
  Berechnung(te, +0.1 / te);
  //200 Sekunden mit konstanter Beschleunigung
  Berechnung(200, 0);
  //innerhalb von 1 Sekunde die Beschleunigung auf 0 senken
  te := 1;
  Berechnung(te, (0 - fDyn1.a) / te);
  //100 Sekunden mit konstanter Geschwindigkeit
  Berechnung(100, 0);

  //Beschleunigungszeit(v0,a1,a2,s_ges:double):double;
  with fdyn1 do
  begin
    a := 0.1;
    //Startdistanz merken oder s auf 0 setzen
    s0 := s;
    //Haltepunkt in s_ges
    //Hier der Anhalteweg bei sofortiger Bremsung
    s_ges :=  (sqr(v) / ((2 * a) * 2));
    // Anhalten bei s = 6000m
    IF 6000-s > s_ges then
      s_ges :=  6000-s;
    memo1.Lines.Add(Format(' Anhalten in %8.2f m ', [s_ges]));
    memo1.Lines.Add('');
    //memo1.Lines.Add(Format('Bremsweg aus v0 %7.2f',[sqr(v)/((2*a)*2)]));
    te := Beschleunigungszeit(v, a, 2 * a, s_ges);
    if te > 0 then
    begin
      Berechnung(te, 0);
      //Reststrecke= Bremsweg
      s0 := s0 + s_ges - s;
      //Bremszeit:t = sqrt(2*s0/a)
      a := -2 * a;
      te := sqrt(-2 * s0 / a);
      Berechnung(te, 0);
    end
    else
      Memo1.Lines.Add('Bremsweg reicht nicht ');
  end;
  Button1.Visible := True;
end;

procedure TFOrm1.Speedtest;
const
  CntEineRichtung = 64;// Zweierpotenzen sind genauer zu rechnen
  runden = 96 * 1024 * 1024 div (CntEineRichtung * 2);

var
  T0, T1: TDateTime;
  i, j, idx: integer;
  function init:Tdyn;
  begin
    with result do
    begin
      t := 0;
      dt := 2 / CntEineRichtung;// Gesamtzeit = 4 Sekunden
      s :=  1;
      v := -2; // v(0) = -v(2) ( v(2) = v0+a*t
      a :=  2;
    end;

  end;

begin
  memo1.Clear;
  memo1.Refresh;
  idx := 0;

  fdyn2:= Init;
  with fdyn2 do
    memo1.Lines.Add(Format(FormStr, [idx, t, s, v, a]));
  //Leerschleife messen
  t0 := now;
  for i := runden - 1 downto 0 do
  begin
    // Einmal vor
    for j := CntEineRichtung - 1 downto 0 do ;
    with fdyn2 do
      a := -a;
    if idx = CntEineRichtung then;
    Inc(idx, CntEineRichtung);
    // und wieder zurück
    for j := CntEineRichtung - 1 downto 0 do ;
    with fdyn2 do
      a := -a;
    Inc(idx, CntEineRichtung);
  end;
  t1 := now - T0;

  fdyn2:= Init;
  idx := 0;
  // Jetzt richtig messen
  t0 := now;
  for i := runden - 1 downto 0 do
  begin
    // Einmal vor
    for j := CntEineRichtung - 1 downto 0 do
      NextTimeStep(Fdyn2, Fdyn2);
    with fdyn2 do
      a := -a;
    Inc(idx, CntEineRichtung);
    if idx = CntEineRichtung then
      with fdyn2 do
        memo1.Lines.Add(Format(FormStr, [idx, t, s, v, a]));
    // und wieder zurück
    for j := CntEineRichtung - 1 downto 0 do
      NextTimeStep(Fdyn2, Fdyn2);
    with fdyn2 do
      a := -a;
    Inc(idx, CntEineRichtung);
  end;
  t1 := now - t1 - t0;

  with fdyn2 do
    memo1.Lines.Add(Format(FormStr, [idx, t, s, v, a]));
  Memo1.Lines.add('Laufzeit Berechnung: '+FormatDateTime('ss.zzz',t1));
  Memo1.Lines.add(Format('Eine Berechnung dauerte %6.2f ns',
                          [T1*86400/(CntEineRichtung*runden*2)* 1E9]));
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  speedtest;
end;

end.
